@@ -1,9 +1,8 @@
-//! Module containing structs related to membership management
 use std::collections::HashMap;
 use std::fmt::Write as FmtWrite;
 use std::io::{Read, Write};
 use std::net::{IpAddr, SocketAddr};
-use std::path::{Path, PathBuf};
+use std::path::PathBuf;
 use std::sync::atomic::{AtomicUsize, Ordering};
 use std::sync::Arc;
 use std::time::Duration;
@@ -12,13 +11,13 @@ use futures::future::join_all;
 use futures::select;
 use futures_util::future::*;
 use serde::{Deserialize, Serialize};
+use tokio::io::AsyncWriteExt;
 use tokio::sync::watch;
 use tokio::sync::Mutex;
 
 use garage_util::background::BackgroundRunner;
 use garage_util::data::*;
 use garage_util::error::Error;
-use garage_util::persister::Persister;
 use garage_util::time::*;
 
 use crate::consul::get_consul_nodes;
@@ -27,36 +26,27 @@ use crate::rpc_client::*;
 use crate::rpc_server::*;
 
 const PING_INTERVAL: Duration = Duration::from_secs(10);
-const DISCOVERY_INTERVAL: Duration = Duration::from_secs(60);
+const CONSUL_INTERVAL: Duration = Duration::from_secs(60);
 const PING_TIMEOUT: Duration = Duration::from_secs(2);
 const MAX_FAILURES_BEFORE_CONSIDERED_DOWN: usize = 5;
 
-/// RPC endpoint used for calls related to membership
 pub const MEMBERSHIP_RPC_PATH: &str = "_membership";
 
-/// RPC messages related to membership
 #[derive(Debug, Serialize, Deserialize)]
 pub enum Message {
-	/// Response to successfull advertisements
 	Ok,
-	/// Message sent to detect other nodes status
 	Ping(PingMessage),
-	/// Ask other node for the nodes it knows. Answered with AdvertiseNodesUp
 	PullStatus,
-	/// Ask other node its config. Answered with AdvertiseConfig
 	PullConfig,
-	/// Advertisement of nodes the host knows up. Sent spontanously or in response to PullStatus
 	AdvertiseNodesUp(Vec<AdvertisedNode>),
-	/// Advertisement of nodes config. Sent spontanously or in response to PullConfig
 	AdvertiseConfig(NetworkConfig),
 }
 
 impl RpcMessage for Message {}
 
-/// A ping, containing informations about status and config
 #[derive(Debug, Serialize, Deserialize)]
 pub struct PingMessage {
-	id: Uuid,
+	id: UUID,
 	rpc_port: u16,
 
 	status_hash: Hash,
@@ -65,29 +55,21 @@ pub struct PingMessage {
 	state_info: StateInfo,
 }
 
-/// A node advertisement
 #[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct AdvertisedNode {
-	/// Id of the node this advertisement relates to
-	pub id: Uuid,
-	/// IP and port of the node
+	pub id: UUID,
 	pub addr: SocketAddr,
 
-	/// Is the node considered up
 	pub is_up: bool,
-	/// When was the node last seen up, in milliseconds since UNIX epoch
 	pub last_seen: u64,
 
 	pub state_info: StateInfo,
 }
 
-/// This node's membership manager
 pub struct System {
-	/// The id of this node
-	pub id: Uuid,
+	pub id: UUID,
 
-	persist_config: Persister<NetworkConfig>,
-	persist_status: Persister<Vec<AdvertisedNode>>,
+	metadata_dir: PathBuf,
 	rpc_local_port: u16,
 
 	state_info: StateInfo,
@@ -95,45 +77,29 @@ pub struct System {
 	rpc_http_client: Arc<RpcHttpClient>,
 	rpc_client: Arc<RpcClient<Message>>,
 
-	replication_factor: usize,
 	pub(crate) status: watch::Receiver<Arc<Status>>,
-	/// The ring
 	pub ring: watch::Receiver<Arc<Ring>>,
 
-	update_lock: Mutex<Updaters>,
+	update_lock: Mutex<(watch::Sender<Arc<Status>>, watch::Sender<Arc<Ring>>)>,
 
-	/// The job runner of this node
 	pub background: Arc<BackgroundRunner>,
 }
 
-struct Updaters {
-	update_status: watch::Sender<Arc<Status>>,
-	update_ring: watch::Sender<Arc<Ring>>,
-}
-
-/// The status of each nodes, viewed by this node
 #[derive(Debug, Clone)]
 pub struct Status {
-	/// Mapping of each node id to its known status
-	pub nodes: HashMap<Uuid, Arc<StatusEntry>>,
-	/// Hash of `nodes`, used to detect when nodes have different views of the cluster
+	pub nodes: HashMap<UUID, Arc<StatusEntry>>,
 	pub hash: Hash,
 }
 
-/// The status of a single node
 #[derive(Debug)]
 pub struct StatusEntry {
-	/// The IP and port used to connect to this node
 	pub addr: SocketAddr,
-	/// Last time this node was seen
 	pub last_seen: u64,
-	/// Number of consecutive pings sent without reply to this node
 	pub num_failures: AtomicUsize,
 	pub state_info: StateInfo,
 }
 
 impl StatusEntry {
-	/// is the node associated to this entry considered up
 	pub fn is_up(&self) -> bool {
 		self.num_failures.load(Ordering::SeqCst) < MAX_FAILURES_BEFORE_CONSIDERED_DOWN
 	}
@@ -178,36 +144,17 @@ impl Status {
 		debug!("END --");
 		self.hash = blake2sum(nodes_txt.as_bytes());
 	}
-
-	fn to_serializable_membership(&self, system: &System) -> Vec<AdvertisedNode> {
-		let mut mem = vec![];
-		for (node, status) in self.nodes.iter() {
-			let state_info = if *node == system.id {
-				system.state_info.clone()
-			} else {
-				status.state_info.clone()
-			};
-			mem.push(AdvertisedNode {
-				id: *node,
-				addr: status.addr,
-				is_up: status.is_up(),
-				last_seen: status.last_seen,
-				state_info,
-			});
-		}
-		mem
-	}
 }
 
-fn gen_node_id(metadata_dir: &Path) -> Result<Uuid, Error> {
-	let mut id_file = metadata_dir.to_path_buf();
+fn gen_node_id(metadata_dir: &PathBuf) -> Result<UUID, Error> {
+	let mut id_file = metadata_dir.clone();
 	id_file.push("node_id");
 	if id_file.as_path().exists() {
 		let mut f = std::fs::File::open(id_file.as_path())?;
 		let mut d = vec![];
 		f.read_to_end(&mut d)?;
 		if d.len() != 32 {
-			return Err(Error::Message("Corrupt node_id file".to_string()));
+			return Err(Error::Message(format!("Corrupt node_id file")));
 		}
 
 		let mut id = [0u8; 32];
@@ -222,42 +169,43 @@ fn gen_node_id(metadata_dir: &Path) -> Result<Uuid, Error> {
 	}
 }
 
+fn read_network_config(metadata_dir: &PathBuf) -> Result<NetworkConfig, Error> {
+	let mut path = metadata_dir.clone();
+	path.push("network_config");
+
+	let mut file = std::fs::OpenOptions::new()
+		.read(true)
+		.open(path.as_path())?;
+
+	let mut net_config_bytes = vec![];
+	file.read_to_end(&mut net_config_bytes)?;
+
+	let net_config = rmp_serde::decode::from_read_ref(&net_config_bytes[..])
+		.expect("Unable to parse network configuration file (has version format changed?).");
+
+	Ok(net_config)
+}
+
 impl System {
-	/// Create this node's membership manager
 	pub fn new(
 		metadata_dir: PathBuf,
 		rpc_http_client: Arc<RpcHttpClient>,
 		background: Arc<BackgroundRunner>,
 		rpc_server: &mut RpcServer,
-		replication_factor: usize,
 	) -> Arc<Self> {
 		let id = gen_node_id(&metadata_dir).expect("Unable to read or generate node ID");
 		info!("Node ID: {}", hex::encode(&id));
 
-		let persist_config = Persister::new(&metadata_dir, "network_config");
-		let persist_status = Persister::new(&metadata_dir, "peer_info");
-
-		let net_config = match persist_config.load() {
+		let net_config = match read_network_config(&metadata_dir) {
 			Ok(x) => x,
 			Err(e) => {
-				match Persister::<garage_rpc_021::ring::NetworkConfig>::new(
-					&metadata_dir,
-					"network_config",
-				)
-				.load()
-				{
-					Ok(old_config) => NetworkConfig::migrate_from_021(old_config),
-					Err(e2) => {
-						info!(
-							"No valid previous network configuration stored ({}, {}), starting fresh.",
-							e, e2
-						);
-						NetworkConfig::new()
-					}
-				}
+				info!(
+					"No valid previous network configuration stored ({}), starting fresh.",
+					e
+				);
+				NetworkConfig::new()
 			}
 		};
-
 		let mut status = Status {
 			nodes: HashMap::new(),
 			hash: Hash::default(),
@@ -268,10 +216,10 @@ impl System {
 		let state_info = StateInfo {
 			hostname: gethostname::gethostname()
 				.into_string()
-				.unwrap_or_else(|_| "<invalid utf-8>".to_string()),
+				.unwrap_or("<invalid utf-8>".to_string()),
 		};
 
-		let ring = Ring::new(net_config, replication_factor);
+		let ring = Ring::new(net_config);
 		let (update_ring, ring) = watch::channel(Arc::new(ring));
 
 		let rpc_path = MEMBERSHIP_RPC_PATH.to_string();
@@ -283,19 +231,14 @@ impl System {
 
 		let sys = Arc::new(System {
 			id,
-			persist_config,
-			persist_status,
+			metadata_dir,
 			rpc_local_port: rpc_server.bind_addr.port(),
 			state_info,
 			rpc_http_client,
 			rpc_client,
-			replication_factor,
 			status,
 			ring,
-			update_lock: Mutex::new(Updaters {
-				update_status,
-				update_ring,
-			}),
+			update_lock: Mutex::new((update_status, update_ring)),
 			background,
 		});
 		sys.clone().register_handler(rpc_server, rpc_path);
@@ -309,18 +252,17 @@ impl System {
 				match msg {
 					Message::Ping(ping) => self2.handle_ping(&addr, &ping).await,
 
-					Message::PullStatus => Ok(self2.handle_pull_status()),
-					Message::PullConfig => Ok(self2.handle_pull_config()),
+					Message::PullStatus => self2.handle_pull_status(),
+					Message::PullConfig => self2.handle_pull_config(),
 					Message::AdvertiseNodesUp(adv) => self2.handle_advertise_nodes_up(&adv).await,
 					Message::AdvertiseConfig(adv) => self2.handle_advertise_config(&adv).await,
 
-					_ => Err(Error::BadRpc("Unexpected RPC message".to_string())),
+					_ => Err(Error::BadRPC(format!("Unexpected RPC message"))),
 				}
 			}
 		});
 	}
 
-	/// Get an RPC client
 	pub fn rpc_client<M: RpcMessage + 'static>(self: &Arc<Self>, path: &str) -> Arc<RpcClient<M>> {
 		RpcClient::new(
 			RpcAddrClient::new(self.rpc_http_client.clone(), path.to_string()),
@@ -329,13 +271,15 @@ impl System {
 		)
 	}
 
-	/// Save network configuration to disc
 	async fn save_network_config(self: Arc<Self>) -> Result<(), Error> {
+		let mut path = self.metadata_dir.clone();
+		path.push("network_config");
+
 		let ring = self.ring.borrow().clone();
-		self.persist_config
-			.save_async(&ring.config)
-			.await
-			.expect("Cannot save current cluster configuration");
+		let data = rmp_to_vec_all_named(&ring.config)?;
+
+		let mut f = tokio::fs::File::create(path.as_path()).await?;
+		f.write_all(&data[..]).await?;
 		Ok(())
 	}
 
@@ -362,27 +306,31 @@ impl System {
 		self.rpc_client.call_many(&to[..], msg, timeout).await;
 	}
 
-	/// Perform bootstraping, starting the ping loop
 	pub async fn bootstrap(
 		self: Arc<Self>,
-		peers: Vec<SocketAddr>,
+		peers: &[SocketAddr],
 		consul_host: Option<String>,
 		consul_service_name: Option<String>,
 	) {
-		let self2 = self.clone();
-		self.background
-			.spawn_worker("discovery loop".to_string(), |stop_signal| {
-				self2.discovery_loop(peers, consul_host, consul_service_name, stop_signal)
-			});
+		let bootstrap_peers = peers.iter().map(|ip| (*ip, None)).collect::<Vec<_>>();
+		self.clone().ping_nodes(bootstrap_peers).await;
 
 		let self2 = self.clone();
 		self.background
-			.spawn_worker("ping loop".to_string(), |stop_signal| {
+			.spawn_worker(format!("ping loop"), |stop_signal| {
 				self2.ping_loop(stop_signal)
 			});
+
+		if let (Some(consul_host), Some(consul_service_name)) = (consul_host, consul_service_name) {
+			let self2 = self.clone();
+			self.background
+				.spawn_worker(format!("Consul loop"), |stop_signal| {
+					self2.consul_loop(stop_signal, consul_host, consul_service_name)
+				});
+		}
 	}
 
-	async fn ping_nodes(self: Arc<Self>, peers: Vec<(SocketAddr, Option<Uuid>)>) {
+	async fn ping_nodes(self: Arc<Self>, peers: Vec<(SocketAddr, Option<UUID>)>) {
 		let ping_msg = self.make_ping();
 		let ping_resps = join_all(peers.iter().map(|(addr, id_option)| {
 			let sys = self.clone();
@@ -430,13 +378,12 @@ impl System {
 				}
 			} else if let Some(id) = id_option {
 				if let Some(st) = status.nodes.get_mut(id) {
-					// we need to increment failure counter as call was done using by_addr so the
-					// counter was not auto-incremented
 					st.num_failures.fetch_add(1, Ordering::SeqCst);
 					if !st.is_up() {
 						warn!("Node {:?} seems to be down.", id);
 						if !ring.config.members.contains_key(id) {
 							info!("Removing node {:?} from status (not in config and not responding to pings anymore)", id);
+							drop(st);
 							status.nodes.remove(&id);
 							has_changes = true;
 						}
@@ -447,10 +394,12 @@ impl System {
 		if has_changes {
 			status.recalculate_hash();
 		}
-		self.update_status(&update_locked, status).await;
+		if let Err(e) = update_locked.0.send(Arc::new(status)) {
+			error!("In ping_nodes: could not save status update ({})", e);
+		}
 		drop(update_locked);
 
-		if !to_advertise.is_empty() {
+		if to_advertise.len() > 0 {
 			self.broadcast(Message::AdvertiseNodesUp(to_advertise), PING_TIMEOUT)
 				.await;
 		}
@@ -471,7 +420,7 @@ impl System {
 		let status_hash = status.hash;
 		let config_version = self.ring.borrow().config.version;
 
-		self.update_status(&update_locked, status).await;
+		update_locked.0.send(Arc::new(status))?;
 		drop(update_locked);
 
 		if is_new || status_hash != ping.status_hash {
@@ -486,13 +435,29 @@ impl System {
 		Ok(self.make_ping())
 	}
 
-	fn handle_pull_status(&self) -> Message {
-		Message::AdvertiseNodesUp(self.status.borrow().to_serializable_membership(self))
+	fn handle_pull_status(&self) -> Result<Message, Error> {
+		let status = self.status.borrow().clone();
+		let mut mem = vec![];
+		for (node, status) in status.nodes.iter() {
+			let state_info = if *node == self.id {
+				self.state_info.clone()
+			} else {
+				status.state_info.clone()
+			};
+			mem.push(AdvertisedNode {
+				id: *node,
+				addr: status.addr,
+				is_up: status.is_up(),
+				last_seen: status.last_seen,
+				state_info,
+			});
+		}
+		Ok(Message::AdvertiseNodesUp(mem))
 	}
 
-	fn handle_pull_config(&self) -> Message {
+	fn handle_pull_config(&self) -> Result<Message, Error> {
 		let ring = self.ring.borrow().clone();
-		Message::AdvertiseConfig(ring.config.clone())
+		Ok(Message::AdvertiseConfig(ring.config.clone()))
 	}
 
 	async fn handle_advertise_nodes_up(
@@ -537,10 +502,10 @@ impl System {
 		if has_changed {
 			status.recalculate_hash();
 		}
-		self.update_status(&update_lock, status).await;
+		update_lock.0.send(Arc::new(status))?;
 		drop(update_lock);
 
-		if !to_ping.is_empty() {
+		if to_ping.len() > 0 {
 			self.background
 				.spawn_cancellable(self.clone().ping_nodes(to_ping).map(Ok));
 		}
@@ -556,8 +521,8 @@ impl System {
 		let ring: Arc<Ring> = self.ring.borrow().clone();
 
 		if adv.version > ring.config.version {
-			let ring = Ring::new(adv.clone(), self.replication_factor);
-			update_lock.update_ring.send(Arc::new(ring))?;
+			let ring = Ring::new(adv.clone());
+			update_lock.1.send(Arc::new(ring))?;
 			drop(update_lock);
 
 			self.background.spawn_cancellable(
@@ -572,7 +537,7 @@ impl System {
 	}
 
 	async fn ping_loop(self: Arc<Self>, mut stop_signal: watch::Receiver<bool>) {
-		while !*stop_signal.borrow() {
+		loop {
 			let restart_at = tokio::time::sleep(PING_INTERVAL);
 
 			let status = self.status.borrow().clone();
@@ -586,74 +551,45 @@ impl System {
 			self.clone().ping_nodes(ping_addrs).await;
 
 			select! {
-				_ = restart_at.fuse() => {},
-				_ = stop_signal.changed().fuse() => {},
-			}
-		}
-	}
-
-	async fn discovery_loop(
-		self: Arc<Self>,
-		bootstrap_peers: Vec<SocketAddr>,
-		consul_host: Option<String>,
-		consul_service_name: Option<String>,
-		mut stop_signal: watch::Receiver<bool>,
-	) {
-		let consul_config = match (consul_host, consul_service_name) {
-			(Some(ch), Some(csn)) => Some((ch, csn)),
-			_ => None,
-		};
-
-		while !*stop_signal.borrow() {
-			let not_configured = self.ring.borrow().config.members.is_empty();
-			let no_peers = self.status.borrow().nodes.len() < 3;
-			let bad_peers = self
-				.status
-				.borrow()
-				.nodes
-				.iter()
-				.filter(|(_, v)| v.is_up())
-				.count() != self.ring.borrow().config.members.len();
-
-			if not_configured || no_peers || bad_peers {
-				info!("Doing a bootstrap/discovery step (not_configured: {}, no_peers: {}, bad_peers: {})", not_configured, no_peers, bad_peers);
-
-				let mut ping_list = bootstrap_peers
-					.iter()
-					.map(|ip| (*ip, None))
-					.collect::<Vec<_>>();
-
-				if let Ok(peers) = self.persist_status.load_async().await {
-					ping_list.extend(peers.iter().map(|x| (x.addr, Some(x.id))));
-				}
-
-				if let Some((consul_host, consul_service_name)) = &consul_config {
-					match get_consul_nodes(consul_host, consul_service_name).await {
-						Ok(node_list) => {
-							ping_list.extend(node_list.iter().map(|a| (*a, None)));
-						}
-						Err(e) => {
-							warn!("Could not retrieve node list from Consul: {}", e);
-						}
+				_ = restart_at.fuse() => (),
+				_ = stop_signal.changed().fuse() => {
+					if *stop_signal.borrow() {
+						return;
 					}
 				}
-
-				self.clone().ping_nodes(ping_list).await;
-			}
-
-			let restart_at = tokio::time::sleep(DISCOVERY_INTERVAL);
-			select! {
-				_ = restart_at.fuse() => {},
-				_ = stop_signal.changed().fuse() => {},
 			}
 		}
 	}
 
-	// for some reason fixing this is causing compilation error, see https://github.com/rust-lang/rust-clippy/issues/7052
-	#[allow(clippy::manual_async_fn)]
+	async fn consul_loop(
+		self: Arc<Self>,
+		mut stop_signal: watch::Receiver<bool>,
+		consul_host: String,
+		consul_service_name: String,
+	) {
+		while !*stop_signal.borrow() {
+			let restart_at = tokio::time::sleep(CONSUL_INTERVAL);
+
+			match get_consul_nodes(&consul_host, &consul_service_name).await {
+				Ok(mut node_list) => {
+					let ping_addrs = node_list.drain(..).map(|a| (a, None)).collect::<Vec<_>>();
+					self.clone().ping_nodes(ping_addrs).await;
+				}
+				Err(e) => {
+					warn!("Could not retrieve node list from Consul: {}", e);
+				}
+			}
+
+			select! {
+				_ = restart_at.fuse() => (),
+				_ = stop_signal.changed().fuse() => (),
+			}
+		}
+	}
+
 	fn pull_status(
 		self: Arc<Self>,
-		peer: Uuid,
+		peer: UUID,
 	) -> impl futures::future::Future<Output = ()> + Send + 'static {
 		async move {
 			let resp = self
@@ -666,7 +602,7 @@ impl System {
 		}
 	}
 
-	async fn pull_config(self: Arc<Self>, peer: Uuid) {
+	async fn pull_config(self: Arc<Self>, peer: UUID) {
 		let resp = self
 			.rpc_client
 			.call(peer, Message::PullConfig, PING_TIMEOUT)
@@ -675,32 +611,4 @@ impl System {
 			let _: Result<_, _> = self.handle_advertise_config(&config).await;
 		}
 	}
-
-	async fn update_status(self: &Arc<Self>, updaters: &Updaters, status: Status) {
-		if status.hash != self.status.borrow().hash {
-			let mut list = status.to_serializable_membership(&self);
-
-			// Combine with old peer list to make sure no peer is lost
-			if let Ok(old_list) = self.persist_status.load_async().await {
-				for pp in old_list {
-					if !list.iter().any(|np| pp.id == np.id) {
-						list.push(pp);
-					}
-				}
-			}
-
-			if !list.is_empty() {
-				info!("Persisting new peer list ({} peers)", list.len());
-				self.persist_status
-					.save_async(&list)
-					.await
-					.expect("Unable to persist peer list");
-			}
-		}
-
-		updaters
-			.update_status
-			.send(Arc::new(status))
-			.expect("Could not update internal membership status");
-	}
 }
